@@ -30,7 +30,8 @@
 // Copyright (C) 2010 OSSD CDAC Mumbai by Leena Chourey (leenac@cdacmumbai.in) and Onkar Potdar (onkar@cdacmumbai.in)
 // Copyright (C) 2011 Joshua Richardson <jric@chegg.com>
 // Copyright (C) 2011 Stephen Reichling <sreichling@chegg.com>
-// Copyright (C) 2012 Igor Slepchin <igor.redhat@gmail.com>
+// Copyright (C) 2011, 2012 Igor Slepchin <igor.slepchin@gmail.com>
+// Copyright (C) 2012 Ihar Filipau <thephilips@gmail.com>
 //
 // To see a description of the changes please see the Changelog file that
 // came with your tarball or type make ChangeLog if you are building from git
@@ -67,13 +68,24 @@
 
 #define DEBUG __FILE__ << ": " << __LINE__ << ": DEBUG: "
 
+class HtmlImage
+{
+public:
+    HtmlImage(GooString *_fName, GfxState *state)
+      : fName(_fName) {
+    state->transform(0, 0, &xMin, &yMax);
+    state->transform(1, 1, &xMax, &yMin);
+  }
+ ~HtmlImage() { delete fName; }
+
+  double xMin, xMax;		// image x coordinates
+  double yMin, yMax;		// image y coordinates
+  GooString  *fName;		// image file name
+};
+
 // returns true if x is closer to y than x is to z
 static inline bool IS_CLOSER(float x, float y, float z) { return fabs((x)-(y)) < fabs((x)-(z)); }
 
-int HtmlPage::pgNum=0;
-int HtmlOutputDev::imgNum=1;
-GooList *HtmlOutputDev::imgList=new GooList();
-
 extern GBool complexMode;
 extern GBool singleHtml;
 extern GBool ignore;
@@ -85,6 +97,8 @@ extern GBool xml;
 extern GBool showHidden;
 extern GBool noMerge;
 
+extern double wordBreakThreshold;
+
 static GBool debug = gFalse;
 static GooString *gstr_buff0 = NULL; // a workspace in which I format strings
 
@@ -160,9 +174,7 @@ HtmlString::HtmlString(GfxState *state, double fontSize, HtmlFontAccu* _fonts) :
     yMax = y - descent * fontSize;
     GfxRGB rgb;
     state->getFillRGB(&rgb);
-    GooString *name = state->getFont()->getName();
-    if (!name) name = HtmlFont::getDefaultFont(); //new GooString("default");
-    HtmlFont hfont=HtmlFont(name, static_cast<int>(fontSize-1), rgb);
+    HtmlFont hfont=HtmlFont(font, static_cast<int>(fontSize-1), rgb);
     if (isMatRotOrSkew(state->getTextMat())) {
       double normalizedMatrix[4];
       memcpy(normalizedMatrix, state->getTextMat(), sizeof(normalizedMatrix));
@@ -258,6 +270,7 @@ HtmlPage::HtmlPage(GBool rawOrder, char *imgExtVal) {
   yxCur1 = yxCur2 = NULL;
   fonts=new HtmlFontAccu();
   links=new HtmlLinks();
+  imgList=new GooList();
   pageWidth=0;
   pageHeight=0;
   fontsPageMarker = 0;
@@ -268,10 +281,11 @@ HtmlPage::HtmlPage(GBool rawOrder, char *imgExtVal) {
 
 HtmlPage::~HtmlPage() {
   clear();
-  if (DocName) delete DocName;
-  if (fonts) delete fonts;
-  if (links) delete links;
-  if (imgExt) delete imgExt;  
+  delete DocName;
+  delete fonts;
+  delete links;
+  delete imgExt;
+  deleteGooList(imgList, HtmlImage);
 }
 
 void HtmlPage::updateFont(GfxState *state) {
@@ -366,7 +380,7 @@ void HtmlPage::addChar(GfxState *state, double x, double y,
       // right, which will not necessarily be the case, e.g. if rotated;
       // It assesses whether or not two characters are close enough to
       // be part of the same string
-      fabs(x1 - curStr->xRight[n-1]) > 0.1 * (curStr->yMax - curStr->yMin) &&
+      fabs(x1 - curStr->xRight[n-1]) > wordBreakThreshold * (curStr->yMax - curStr->yMin) &&
       // rotation is (cos q, sin q, -sin q, cos q, 0, 0)
       // sin q is zero iff there is no rotation, or 180 deg. rotation;
       // for 180 rotation, cos q will be negative
@@ -385,7 +399,19 @@ void HtmlPage::addChar(GfxState *state, double x, double y,
     h1 /= uLen;
   }
   for (i = 0; i < uLen; ++i) {
-    curStr->addChar(state, x1 + i*w1, y1 + i*h1, w1, h1, u[i]);
+    Unicode u1 = u[i];
+    if (u1 >= 0xd800 && u1 <= 0xdbff && i < uLen) {
+      // surrogate pair
+      const Unicode u2 = u[i + 1];
+      if (u2 >= 0xdc00 && u2 <= 0xdfff) {
+	u1 = 0x10000 + ((u1 - 0xd800) << 10) + (u2 - 0xdc00);
+	
+	curStr->addChar(state, x1 + i*w1, y1 + i*h1, w1, h1, u1);
+      }
+      ++i;
+    } else {
+      curStr->addChar(state, x1 + i*w1, y1 + i*h1, w1, h1, u1);
+    }
   }
 }
 
@@ -600,7 +626,7 @@ void HtmlPage::coalesce() {
     {
 //      printf("yes\n");
       n = str1->len + str2->len;
-      if ((addSpace = horSpace > 0.1 * space)) {
+      if ((addSpace = horSpace > wordBreakThreshold * space)) {
         ++n;
       }
       if (addLineBreak) {
@@ -728,6 +754,15 @@ void HtmlPage::dumpAsXML(FILE* f,int page){
     delete fontCSStyle;
   }
   
+  int listlen=imgList->getLength();
+  for (int i = 0; i < listlen; i++) {
+    HtmlImage *img = (HtmlImage*)imgList->del(0);
+    fprintf(f,"<image top=\"%d\" left=\"%d\" ",xoutRound(img->yMin),xoutRound(img->xMin));
+    fprintf(f,"width=\"%d\" height=\"%d\" ",xoutRound(img->xMax-img->xMin),xoutRound(img->yMax-img->yMin));
+    fprintf(f,"src=\"%s\"/>\n",img->fName->getCString());
+    delete img;
+  }
+
   for(HtmlString *tmp=yxStrings;tmp;tmp=tmp->yxNext){
     if (tmp->htext){
       fprintf(f,"<text top=\"%d\" left=\"%d\" ",xoutRound(tmp->yMin),xoutRound(tmp->xMin));
@@ -740,6 +775,42 @@ void HtmlPage::dumpAsXML(FILE* f,int page){
   fputs("</page>\n",f);
 }
 
+static void printCSS(FILE *f)
+{
+  // Image flip/flop CSS
+  // Source:
+  // http://stackoverflow.com/questions/1309055/cross-browser-way-to-flip-html-image-via-javascript-css
+  // tested in Chrome, Fx (Linux) and IE9 (W7)
+  static const char css[] = 
+    "<STYLE type=\"text/css\">" "\n"
+    "<!--" "\n"
+    ".xflip {" "\n"
+    "    -moz-transform: scaleX(-1);" "\n"
+    "    -webkit-transform: scaleX(-1);" "\n"
+    "    -o-transform: scaleX(-1);" "\n"
+    "    transform: scaleX(-1);" "\n"
+    "    filter: fliph;" "\n"
+    "}" "\n"
+    ".yflip {" "\n"
+    "    -moz-transform: scaleY(-1);" "\n"
+    "    -webkit-transform: scaleY(-1);" "\n"
+    "    -o-transform: scaleY(-1);" "\n"
+    "    transform: scaleY(-1);" "\n"
+    "    filter: flipv;" "\n"
+    "}" "\n"
+    ".xyflip {" "\n"
+    "    -moz-transform: scaleX(-1) scaleY(-1);" "\n"
+    "    -webkit-transform: scaleX(-1) scaleY(-1);" "\n"
+    "    -o-transform: scaleX(-1) scaleY(-1);" "\n"
+    "    transform: scaleX(-1) scaleY(-1);" "\n"
+    "    filter: fliph + flipv;" "\n"
+    "}" "\n"
+    "-->" "\n"
+    "</STYLE>" "\n";
+
+  fwrite( css, sizeof(css)-1, 1, f );
+}
+
 int HtmlPage::dumpComplexHeaders(FILE * const file, FILE *& pageFile, int page) {
   GooString* tmp;
 
@@ -756,7 +827,7 @@ int HtmlPage::dumpComplexHeaders(FILE * const file, FILE *& pageFile, int page)
       }
       delete pgNum;
       if (!pageFile) {
-	  error(-1, "Couldn't open html file '%s'", tmp->getCString());
+	  error(errIO, -1, "Couldn't open html file '{0:t}'", tmp);
 	  delete tmp;
 	  return 1;
       } 
@@ -791,7 +862,7 @@ void HtmlPage::dumpComplex(FILE *file, int page){
 
   if( firstPage == -1 ) firstPage = page; 
   
-  if (dumpComplexHeaders(file, pageFile, page)) { error(-1, "Couldn't write headers."); return; }
+  if (dumpComplexHeaders(file, pageFile, page)) { error(errIO, -1, "Couldn't write headers."); return; }
 
   tmp=basename(DocName);
    
@@ -865,13 +936,19 @@ void HtmlPage::dump(FILE *f, int pageNum)
   {
     fprintf(f,"<A name=%d></a>",pageNum);
     // Loop over the list of image names on this page
-    int listlen=HtmlOutputDev::imgList->getLength();
+    int listlen=imgList->getLength();
     for (int i = 0; i < listlen; i++) {
-      GooString *fName= (GooString *)HtmlOutputDev::imgList->del(0);
-      fprintf(f,"<IMG src=\"%s\"/><br/>\n",fName->getCString());
-      delete fName;
+      HtmlImage *img = (HtmlImage*)imgList->del(0);
+
+      // see printCSS() for class names
+      const char *styles[4] = { "", " class=\"xflip\"", " class=\"yflip\"", " class=\"xyflip\"" };
+      int style_index=0;
+      if (img->xMin > img->xMax) style_index += 1; // xFlip
+      if (img->yMin > img->yMax) style_index += 2; // yFlip
+
+      fprintf(f,"<IMG%s src=\"%s\"/><br/>\n",styles[style_index],img->fName->getCString());
+      delete img;
     }
-    HtmlOutputDev::imgNum=1;
 
     GooString* str;
     for(HtmlString *tmp=yxStrings;tmp;tmp=tmp->yxNext){
@@ -924,11 +1001,16 @@ void HtmlPage::setDocName(char *fname){
   DocName=new GooString(fname);
 }
 
+void HtmlPage::addImage(GooString *fname, GfxState *state) {
+  HtmlImage *img = new HtmlImage(fname, state);
+  imgList->append(img);
+}
+
 //------------------------------------------------------------------------
 // HtmlMetaVar
 //------------------------------------------------------------------------
 
-HtmlMetaVar::HtmlMetaVar(char *_name, char *_content)
+HtmlMetaVar::HtmlMetaVar(const char *_name, const char *_content)
 {
     name = new GooString(_name);
     content = new GooString(_content);
@@ -954,7 +1036,7 @@ GooString* HtmlMetaVar::toString()
 // HtmlOutputDev
 //------------------------------------------------------------------------
 
-static char* HtmlEncodings[][2] = {
+static const char* HtmlEncodings[][2] = {
     {"Latin1", "ISO-8859-1"},
     {NULL, NULL}
 };
@@ -979,7 +1061,7 @@ void HtmlOutputDev::doFrame(int firstPage){
   fName->append(".html");
 
   if (!(fContentsFrame = fopen(fName->getCString(), "w"))){
-    error(-1, "Couldn't open html file '%s'", fName->getCString());
+    error(errIO, -1, "Couldn't open html file '{0:t}'", fName);
     delete fName;
     return;
   }
@@ -1010,11 +1092,12 @@ void HtmlOutputDev::doFrame(int firstPage){
   fclose(fContentsFrame);  
 }
 
-HtmlOutputDev::HtmlOutputDev(char *fileName, char *title, 
+HtmlOutputDev::HtmlOutputDev(Catalog *catalogA, char *fileName, char *title, 
 	char *author, char *keywords, char *subject, char *date,
 	char *extension,
 	GBool rawOrder, int firstPage, GBool outline) 
 {
+  catalog = catalogA;
   fContentsFrame = NULL;
   docTitle = new GooString(title);
   pages = NULL;
@@ -1023,7 +1106,6 @@ HtmlOutputDev::HtmlOutputDev(char *fileName, char *title,
   this->rawOrder = rawOrder;
   this->doOutline = outline;
   ok = gFalse;
-  imgNum=1;
   //this->firstPage = firstPage;
   //pageNum=firstPage;
   // open file
@@ -1055,7 +1137,7 @@ HtmlOutputDev::HtmlOutputDev(char *fileName, char *title,
 
          if (!(fContentsFrame = fopen(left->getCString(), "w")))
          {
-             error(-1, "Couldn't open html file '%s'", left->getCString());
+             error(errIO, -1, "Couldn't open html file '{0:t}'", left);
              delete left;
              return;
          }
@@ -1077,13 +1159,15 @@ HtmlOutputDev::HtmlOutputDev(char *fileName, char *title,
        right->append("s.html");
 
        if (!(page=fopen(right->getCString(),"w"))){
-        error(-1, "Couldn't open html file '%s'", right->getCString());
+        error(errIO, -1, "Couldn't open html file '{0:t}'", right);
         delete right;
 		return;
        }
        delete right;
        fputs(DOCTYPE, page);
-       fputs("<HTML>\n<HEAD>\n<TITLE></TITLE>\n</HEAD>\n<BODY>\n",page);
+       fputs("<HTML>\n<HEAD>\n<TITLE></TITLE>\n",page);
+       printCSS(page);
+       fputs("</HEAD>\n<BODY>\n",page);
      }
   }
 
@@ -1094,7 +1178,7 @@ HtmlOutputDev::HtmlOutputDev(char *fileName, char *title,
       if (!xml) right->append(".html");
       if (xml) right->append(".xml");
       if (!(page=fopen(right->getCString(),"w"))){
-	error(-1, "Couldn't open html file '%s'", right->getCString());
+	error(errIO, -1, "Couldn't open html file '{0:t}'", right);
 	delete right;
 	return;
       }  
@@ -1106,7 +1190,7 @@ HtmlOutputDev::HtmlOutputDev(char *fileName, char *title,
     {
       fprintf(page, "<?xml version=\"1.0\" encoding=\"%s\"?>\n", htmlEncoding->getCString());
       fputs("<!DOCTYPE pdf2xml SYSTEM \"pdf2xml.dtd\">\n\n", page);
-      fputs("<pdf2xml>\n",page);
+      fprintf(page,"<pdf2xml producer=\"%s\" version=\"%s\">\n", PACKAGE_NAME, PACKAGE_VERSION);
     } 
     else 
     {
@@ -1115,6 +1199,7 @@ HtmlOutputDev::HtmlOutputDev(char *fileName, char *title,
       fprintf(page, "<META http-equiv=\"Content-Type\" content=\"text/html; charset=%s\"/>\n", htmlEncoding->getCString());
       
       dumpMetaVars(page);
+      printCSS(page);
       fprintf(page,"</HEAD>\n");
       fprintf(page,"<BODY bgcolor=\"#A0A0A0\" vlink=\"blue\" link=\"blue\">\n");
     }
@@ -1135,14 +1220,16 @@ HtmlOutputDev::~HtmlOutputDev() {
       fputs("</BODY>\n</HTML>\n",fContentsFrame);  
       fclose(fContentsFrame);
     }
-    if (xml) {
-      fputs("</pdf2xml>\n",page);  
-      fclose(page);
-    } else
-    if ( !complexMode || xml || noframes )
-    { 
-      fputs("</BODY>\n</HTML>\n",page);  
-      fclose(page);
+    if (page != NULL) {
+      if (xml) {
+        fputs("</pdf2xml>\n",page);  
+        fclose(page);
+      } else
+      if ( !complexMode || xml || noframes )
+      { 
+        fputs("</BODY>\n</HTML>\n",page);  
+        fclose(page);
+      }
     }
     if (pages)
       delete pages;
@@ -1191,7 +1278,7 @@ void HtmlOutputDev::startPage(int pageNum, GfxState *state) {
 
 
 void HtmlOutputDev::endPage() {
-  Links *linksList = docPage->getLinks(catalog);
+  Links *linksList = docPage->getLinks();
   for (int i = 0; i < linksList->getNumLinks(); ++i)
   {
       doProcessLink(linksList->getLink(i));
@@ -1235,133 +1322,68 @@ void HtmlOutputDev::drawChar(GfxState *state, double x, double y,
   pages->addChar(state, x, y, dx, dy, originX, originY, u, uLen);
 }
 
-void HtmlOutputDev::drawImageMask(GfxState *state, Object *ref, Stream *str,
-				  int width, int height, GBool invert,
-				  GBool interpolate, GBool inlineImg) {
-
-  if (ignore||complexMode) {
-    OutputDev::drawImageMask(state, ref, str, width, height, invert, interpolate, inlineImg);
-    return;
-  }
-  
+void HtmlOutputDev::drawJpegImage(GfxState *state, Stream *str)
+{
   FILE *f1;
   int c;
-  
-  // dump JPEG file
-  if (dumpJPEG  && str->getKind() == strDCT) {
-    GooString *fName=new GooString(Docname);
-    fName->append("-");
-    GooString *pgNum=GooString::fromInt(pageNum);
-    GooString *imgnum=GooString::fromInt(imgNum);
-    // open the image file
-    fName->append(pgNum)->append("_")->append(imgnum)->append(".jpg");
-    delete pgNum;
-    delete imgnum;
 
-    ++imgNum;
-    if (!(f1 = fopen(fName->getCString(), "wb"))) {
-      error(-1, "Couldn't open image file '%s'", fName->getCString());
-      delete fName;
-      return;
-    }
-
-    // initialize stream
-    str = str->getNextStream();
-    str->reset();
-
-    // copy the stream
-    while ((c = str->getChar()) != EOF)
-      fputc(c, f1);
-
-    fclose(f1);
-   
-  if (fName) imgList->append(fName);
+  // open the image file
+  GooString *fName=createImageFileName("jpg");
+  if (!(f1 = fopen(fName->getCString(), "wb"))) {
+    error(errIO, -1, "Couldn't open image file '%s'", fName->getCString());
+    delete fName;
+    return;
   }
-  else {
-    OutputDev::drawImageMask(state, ref, str, width, height, invert, interpolate, inlineImg);
+
+  // initialize stream
+  str = str->getNextStream();
+  str->reset();
+
+  // copy the stream
+  while ((c = str->getChar()) != EOF)
+    fputc(c, f1);
+
+  fclose(f1);
+
+  if (fName) {
+      pages->addImage(fName, state);
   }
 }
 
-void HtmlOutputDev::drawImage(GfxState *state, Object *ref, Stream *str,
-			      int width, int height, GfxImageColorMap *colorMap,
-			      GBool interpolate, int *maskColors, GBool inlineImg) {
+void HtmlOutputDev::drawPngImage(GfxState *state, Stream *str, int width, int height,
+                                 GfxImageColorMap *colorMap, GBool isMask)
+{
+#ifdef ENABLE_LIBPNG
+  FILE *f1;
 
-  if (ignore||complexMode) {
-    OutputDev::drawImage(state, ref, str, width, height, colorMap, interpolate,
-			 maskColors, inlineImg);
+  if (!colorMap && !isMask) {
+    error(errInternal, -1, "Can't have color image without a color map");
     return;
   }
 
-  FILE *f1;
-  int c;
-  
-  /*if( !globalParams->getErrQuiet() )
-    printf("image stream of kind %d\n", str->getKind());*/
-  // dump JPEG file
-  if (dumpJPEG && str->getKind() == strDCT) {
-    GooString *fName=new GooString(Docname);
-    fName->append("-");
-    GooString *pgNum= GooString::fromInt(pageNum);
-    GooString *imgnum= GooString::fromInt(imgNum);  
-    
-    // open the image file
-    fName->append(pgNum)->append("_")->append(imgnum)->append(".jpg");
-    delete pgNum;
-    delete imgnum;
-
-    ++imgNum;
-    
-    if (!(f1 = fopen(fName->getCString(), "wb"))) {
-      error(-1, "Couldn't open image file '%s'", fName->getCString());
-      delete fName;
-      return;
-    }
-
-    // initialize stream
-    str = str->getNextStream();
-    str->reset();
-
-    // copy the stream
-    while ((c = str->getChar()) != EOF)
-      fputc(c, f1);
-    
-    fclose(f1);
-  
-    if (fName) imgList->append(fName);
+  // open the image file
+  GooString *fName=createImageFileName("png");
+  if (!(f1 = fopen(fName->getCString(), "wb"))) {
+    error(errIO, -1, "Couldn't open image file '%s'", fName->getCString());
+    delete fName;
+    return;
   }
-  else {
-#ifdef ENABLE_LIBPNG
-    // Dump the image as a PNG file. Much of the PNG code
-    // comes from an example by Guillaume Cottenceau.
+
+  PNGWriter *writer = new PNGWriter( isMask ? PNGWriter::MONOCHROME : PNGWriter::RGB );
+  // TODO can we calculate the resolution of the image?
+  if (!writer->init(f1, width, height, 72, 72)) {
+    error(errInternal, -1, "Can't init PNG for image '%s'", fName->getCString());
+    delete writer;
+    fclose(f1);
+    return;
+  }
+
+  if (!isMask) {
     Guchar *p;
     GfxRGB rgb;
-    png_byte *row = (png_byte *) malloc(3 * width);   // 3 bytes/pixel: RGB
+    png_byte *row = (png_byte *) gmalloc(3 * width);   // 3 bytes/pixel: RGB
     png_bytep *row_pointer= &row;
 
-    // Create the image filename
-    GooString *fName=new GooString(Docname);
-    fName->append("-");
-    GooString *pgNum= GooString::fromInt(pageNum);
-    GooString *imgnum= GooString::fromInt(imgNum);  
-    fName->append(pgNum)->append("_")->append(imgnum)->append(".png");
-    delete pgNum;
-    delete imgnum;
-
-    // Open the image file
-    if (!(f1 = fopen(fName->getCString(), "wb"))) {
-      error(-1, "Couldn't open image file '%s'", fName->getCString());
-      delete fName;
-      return;
-    }
-
-    PNGWriter *writer = new PNGWriter();
-    // TODO can we calculate the resolution of the image?
-    if (!writer->init(f1, width, height, 72, 72)) {
-        delete writer;
-        fclose(f1);
-        return;
-    }
-
     // Initialize the image stream
     ImageStream *imgStr = new ImageStream(str, width,
                         colorMap->getNumPixelComps(), colorMap->getBits());
@@ -1374,32 +1396,125 @@ void HtmlOutputDev::drawImage(GfxState *state, Object *ref, Stream *str,
       p = imgStr->getLine();
       for (int x = 0; x < width; x++) {
         colorMap->getRGB(p, &rgb);
-	// Write the RGB pixels into the row
-	row[3*x]= colToByte(rgb.r);
-	row[3*x+1]= colToByte(rgb.g);
-	row[3*x+2]= colToByte(rgb.b);
-         p += colorMap->getNumPixelComps();
+        // Write the RGB pixels into the row
+        row[3*x]= colToByte(rgb.r);
+        row[3*x+1]= colToByte(rgb.g);
+        row[3*x+2]= colToByte(rgb.b);
+        p += colorMap->getNumPixelComps();
       }
 
       if (!writer->writeRow(row_pointer)) {
+        error(errIO, -1, "Failed to write into PNG '%s'", fName->getCString());
         delete writer;
+        delete imgStr;
         fclose(f1);
         return;
       }
     }
-
-    writer->close();
-    delete writer;
-    fclose(f1);
-
-    free(row);
-    imgList->append(fName);
-    ++imgNum;
+    gfree(row);
     imgStr->close();
     delete imgStr;
+  }
+  else { // isMask == true
+    ImageStream *imgStr = new ImageStream(str, width, 1, 1);
+    imgStr->reset();
+
+    Guchar *png_row = (Guchar *)gmalloc( width );
+
+    for (int ri = 0; ri < height; ++ri)
+    {
+      // read the row of the mask
+      Guchar *bit_row = imgStr->getLine();
+
+      // invert for PNG
+      for(int i = 0; i < width; i++)
+        png_row[i] = bit_row[i] ? 0xff : 0x00;
+
+      if (!writer->writeRow( &png_row ))
+      {
+        error(errIO, -1, "Failed to write into PNG '%s'", fName->getCString());
+        delete writer;
+        fclose(f1);
+        delete imgStr;
+        gfree(png_row);
+        return;
+      }
+    }
+    imgStr->close();
+    delete imgStr;
+    gfree(png_row);
+  }
+
+  str->close();
+
+  writer->close();
+  delete writer;
+  fclose(f1);
+
+  pages->addImage(fName, state);
 #else
+  return;
+#endif
+}
+
+GooString *HtmlOutputDev::createImageFileName(const char *ext)
+{
+  GooString *fName=new GooString(Docname);
+  fName->append("-");
+  GooString *pgNum= GooString::fromInt(pageNum);
+  GooString *imgnum= GooString::fromInt(pages->getNumImages()+1);
+
+  fName->append(pgNum)->append("_")->append(imgnum)->append(".")->append(ext);
+  delete pgNum;
+  delete imgnum;
+
+  return fName;
+}
+
+void HtmlOutputDev::drawImageMask(GfxState *state, Object *ref, Stream *str,
+				  int width, int height, GBool invert,
+				  GBool interpolate, GBool inlineImg) {
+
+  if (ignore||(complexMode && !xml)) {
+    OutputDev::drawImageMask(state, ref, str, width, height, invert, interpolate, inlineImg);
+    return;
+  }
+  
+  // dump JPEG file
+  if (dumpJPEG  && str->getKind() == strDCT) {
+    drawJpegImage(state, str);
+  }
+  else {
+#ifdef ENABLE_LIBPNG
+    drawPngImage(state, str, width, height, NULL, gTrue);
+#else
+    OutputDev::drawImageMask(state, ref, str, width, height, invert, interpolate, inlineImg);
+#endif
+  }
+}
+
+void HtmlOutputDev::drawImage(GfxState *state, Object *ref, Stream *str,
+			      int width, int height, GfxImageColorMap *colorMap,
+			      GBool interpolate, int *maskColors, GBool inlineImg) {
+
+  if (ignore||(complexMode && !xml)) {
     OutputDev::drawImage(state, ref, str, width, height, colorMap, interpolate,
 			 maskColors, inlineImg);
+    return;
+  }
+  
+  /*if( !globalParams->getErrQuiet() )
+    printf("image stream of kind %d\n", str->getKind());*/
+  // dump JPEG file
+  if (dumpJPEG && str->getKind() == strDCT) {
+    drawJpegImage(state, str);
+  }
+  else {
+#ifdef ENABLE_LIBPNG
+    drawPngImage(state, str, width, height, colorMap );
+#else
+    OutputDev::drawImage(state, ref, str, width, height, colorMap, interpolate,
+                         maskColors, inlineImg);
 #endif
   }
 }
@@ -1416,13 +1531,13 @@ void HtmlOutputDev::doProcessLink(AnnotLink* link){
   cvtUserToDev(_x2,_y2,&x2,&y2); 
 
 
-  GooString* _dest=getLinkDest(link,catalog);
+  GooString* _dest=getLinkDest(link);
   HtmlLink t((double) x1,(double) y2,(double) x2,(double) y1,_dest);
   pages->AddLink(t);
   delete _dest;
 }
 
-GooString* HtmlOutputDev::getLinkDest(AnnotLink *link,Catalog* catalog){
+GooString* HtmlOutputDev::getLinkDest(AnnotLink *link){
   char *p;
   if (!link->getAction())
     return new GooString();
@@ -1563,7 +1678,7 @@ GBool HtmlOutputDev::dumpDocOutline(PDFDoc* doc)
 	GBool bClose = gFalse;
 	Catalog *catalog = doc->getCatalog();
 
-	if (!ok || xml)
+	if (!ok)
                 return gFalse;
   
 	Outline *outline = doc->getOutline();
@@ -1574,7 +1689,7 @@ GBool HtmlOutputDev::dumpDocOutline(PDFDoc* doc)
 	if (!outlines)
 		return gFalse;
   
-	if (!complexMode && !xml)
+	if (!complexMode || xml)
   	{
 		output = page;
   	}
@@ -1609,21 +1724,30 @@ GBool HtmlOutputDev::dumpDocOutline(PDFDoc* doc)
 		}
 	}
  
-  	GBool done = newOutlineLevel(output, outlines, catalog);
-  	if (done && !complexMode)
-    	fputs("<hr>\n", output);
-	
-	if (bClose)
+	if (!xml)
 	{
-		fputs("</BODY>\n</HTML>\n", output);
-		fclose(output);
+		GBool done = newHtmlOutlineLevel(output, outlines, catalog);
+		if (done && !complexMode)
+			fputs("<hr>\n", output);
+	
+		if (bClose)
+		{
+			fputs("</BODY>\n</HTML>\n", output);
+			fclose(output);
+		}
 	}
-  	return done;
+	else
+		newXmlOutlineLevel(output, outlines, catalog);
+
+	return gTrue;
 #endif
 }
 
-GBool HtmlOutputDev::newOutlineLevel(FILE *output, GooList *outlines, Catalog* catalog, int level)
+GBool HtmlOutputDev::newHtmlOutlineLevel(FILE *output, GooList *outlines, Catalog* catalog, int level)
 {
+#ifdef DISABLE_OUTLINE
+	return gFalse;
+#else
 	GBool atLeastOne = gFalse;
 
 	if (level == 1)
@@ -1639,30 +1763,10 @@ GBool HtmlOutputDev::newOutlineLevel(FILE *output, GooList *outlines, Catalog* c
 		GooString *titleStr = HtmlFont::HtmlFilter(item->getTitle(),
 							   item->getTitleLength());
 
-		// get corresponding link
 		GooString *linkName = NULL;;
-		LinkAction *action = item->getAction();
-		LinkGoTo *link = NULL;
-		Object dest;
-		if (action && action->getKind() == actionGoTo)
-			link = dynamic_cast<LinkGoTo*>(action);
-		if (link && link->isOk()) {
-			LinkDest *linkdest=NULL;
-			if (link->getDest()!=NULL)
-				linkdest=link->getDest()->copy();
-			else if (link->getNamedDest()!=NULL)
-				linkdest=catalog->findDest(link->getNamedDest());
-
-			if (linkdest) {
-				int page;
-				if (linkdest->isPageRef()) {
-					Ref pageref=linkdest->getPageRef();
-					page=catalog->findPage(pageref.num,pageref.gen);
-				} else {
-					page=linkdest->getPageNum();
-				}
-				delete linkdest;
-
+        int page = getOutlinePageNum(item);
+        if (page > 0)
+        {
 				/*		complex		simple
 				frames		file-4.html	files.html#4
 				noframes	file.html#4	file.html#4
@@ -1683,9 +1787,7 @@ GBool HtmlOutputDev::newOutlineLevel(FILE *output, GooList *outlines, Catalog* c
 					}
 				}
 				delete str;
-			}
 		}
-		dest.free();
 
 		fputs("<li>",output);
 		if (linkName)
@@ -1702,7 +1804,7 @@ GBool HtmlOutputDev::newOutlineLevel(FILE *output, GooList *outlines, Catalog* c
 		if (item->hasKids())
 		{
 			fputs("\n",output);
-			newOutlineLevel(output, item->getKids(), catalog, level+1);
+			newHtmlOutlineLevel(output, item->getKids(), catalog, level+1);
 		}
 		item->close();
 		fputs("</li>\n",output);
@@ -1710,4 +1812,75 @@ GBool HtmlOutputDev::newOutlineLevel(FILE *output, GooList *outlines, Catalog* c
 	fputs("</ul>\n",output);
 
 	return atLeastOne;
+#endif
 }
+
+void HtmlOutputDev::newXmlOutlineLevel(FILE *output, GooList *outlines, Catalog* catalog)
+{
+#ifndef DISABLE_OUTLINE
+    fputs("<outline>\n", output);
+
+    for (int i = 0; i < outlines->getLength(); i++)
+    {
+        OutlineItem *item     = (OutlineItem*)outlines->get(i);
+        GooString   *titleStr = HtmlFont::HtmlFilter(item->getTitle(),
+                                                     item->getTitleLength());
+        int page = getOutlinePageNum(item);
+        if (page > 0)
+        {
+            fprintf(output, "<item page=\"%d\">%s</item>\n",
+                    page, titleStr->getCString());
+        }
+        else
+        {
+            fprintf(output, "<item>%s</item>\n", titleStr->getCString());
+        }
+        delete titleStr;
+
+        item->open();
+        if (item->hasKids())
+        {
+            newXmlOutlineLevel(output, item->getKids(), catalog);
+        }
+        item->close();
+    }    
+
+    fputs("</outline>\n", output);
+#endif
+}
+
+#ifndef DISABLE_OUTLINE
+int HtmlOutputDev::getOutlinePageNum(OutlineItem *item)
+{
+    LinkAction *action   = item->getAction();
+    LinkGoTo   *link     = NULL;
+    LinkDest   *linkdest = NULL;
+    int         pagenum  = -1;
+
+    if (!action || action->getKind() != actionGoTo)
+        return pagenum;
+
+    link = dynamic_cast<LinkGoTo*>(action);
+
+    if (!link || !link->isOk())
+        return pagenum;
+
+    if (link->getDest())
+        linkdest = link->getDest()->copy();
+    else if (link->getNamedDest())
+        linkdest = catalog->findDest(link->getNamedDest());
+
+    if (!linkdest)
+        return pagenum;
+
+    if (linkdest->isPageRef()) {
+        Ref pageref = linkdest->getPageRef();
+        pagenum = catalog->findPage(pageref.num, pageref.gen);
+    } else {
+        pagenum = linkdest->getPageNum();
+    }
+
+    delete linkdest;
+    return pagenum;
+}
+#endif
